#include <xc.h>
#include <stdlib.h>
#define bversion 0 // =1 for green and 'b' red display version, =0 for other
/*
Hackaday badge greyscale display demos
Mike Harrison www.electricstuff.co.uk

Implements greyscale display vis binary code modulation.
uses simple squaring gamma correction for nicer greyscale (displayed on-time = pixel value ^2)

compiled with XC8 V1.20 in Standard mode
May behave differently in free mode,in particular t1offset may need adjusting to get 
monotonic greyscale. 

Doesn't make use of kernel but should tolerate it.
Compile with offset 0x1820.
needs this file and sintable.inc
Includes vector redirects if programmed directly into a device without bootloader

Sorry, effects are nasty and not documented but hey, it's meant to be a hack isn't it?

*/


#define vflip 1 // displayorientation =1 for 0,0 to be top-left, -0 for bot-left

#pragma config config0, FOSC=INTOSCIO, LS48MHZ=SYS48X8, LS48MHZ=SYS48X8, PLLSEL=PLL3X,CFGPLLEN=1,BOREN=OFF,PCLKEN=OFF,PBADEN=OFF,WDTEN=OFF



#define lbut 1 // button flags in buttonstatea
#define rbut 8
#define tbut 2
#define bbut 4

#define offtime 2 // rames button held till turnoff

#define left (buttons&lbut)
#define right (buttons&rbut)
#define up (buttons & tbut)
#define down (buttons&bbut)

#define leftpress (butpress & lbut)
#define rightpress (butpress  & rbut)
#define uppress (butpress & tbut)
#define downpress (butpress & bbut)

#define topbut PORTBbits.RB0

//______________  Greyscale display defines & globals

#define nbits 8 // greyscale display bits per pixel - NB only 8 now supported - >8 only worthwhile on higher-res white display
                
#define bcmtime 16 // length of shortest BCM 'on' pulse, cycles. row time = bcmtime*2^nbits, plus deadtime overheads
#define bitsinint 3 // how many short bits to do inside first int on row. too low : timing will be wrong, too high : increased interrupt load
                    // optimum value depends on bcmtime

#define t1offset 29  // cycles taken going in & out of int code to correct timer1 int timing - value may differ for differnt compler versions
#define spibufsize nbits*8*16/8  // buffer for BCM data 
#define imgbufsize (16*8) // greyscale image buffer
#define presquare 1 // =1 to precalculate gamma square data - faster but uses more RAM & ROM Shuffle time = 7mS / 21mS for presquare = 1/0

 #if presquare==1
unsigned char sqbuf[imgbufsize]; // buffer for precalculated gamma-corrected pixel data, to avoid calculating for each bit
#endif

unsigned char imgbuf[8*16], spibuf[2][spibufsize];

// flags & vars used by int code
volatile unsigned char disppage=0,newpage=0,frametimer,buttons; 

//________________________________________________
//misc globals 

unsigned char lastbut,butpress,speed;
 unsigned char effect;  
bit start;

//______________________________ vector bodging
#define hadvec_reset 0x180c // relocated vectors from HADbadge kernel
#define hadvec_loint 0x1806
#define hadvec_hiint 0x1800

#define codeoffset 0x1820 // codeoffset value given to linker

void vecs(void)@0x00
{
#asm
org 0x00
// redirects for when programmed direct into device with no bootloader
goto codeoffset
org 0x08 //hipri int
goto codeoffset+8
org 0x18 //lopri int
goto codeoffset+0x18

//fixups for  vector addresses used in HADbadge kernel/bootloader

org hadvec_hiint // redirected lopri int vector from bootloader 
goto codeoffset+0x08

org hadvec_loint // hipri int vector from bootloader
goto codeoffset+0x18

org hadvec_reset// reset vector 
goto codeoffset
#endasm

}


//____________________________________________________________________ Timer 1 interrupt for greyscale display
/*
  sends pre-formatted data from spibuf, and enables display for appropriate time for that bit.   
  general principle: 
  enable columns on all pixels with bit 0 set, turn row on for time <t>
  enable columns on all pixels with bit 1 set, turn row on for time <t*2> 
  etc.
  
  repeat for each row. 

As you may want to have an update framerate slower than the display refresh rate, the formatted data 
(16 column bits of on/off for each bit of greyscale) is precalculated by 'shuffle' procedure, so interrupt
 task only has to send the SPI data and do the pulses. 
As the range of pulse widths is large, to save interrupt overheads the shortest ones are done entirely inside the 
first interrupt in each row to save the call/return overhead.
Longer pulses are done using the timer interrupt  
Data is double-buffered as a complete frame must be displayed before updating to avoid sparkly artefacts due to
incorrect pulse lengths in a truncated output frame.

Pulses would ideally be generated using PWM or compare hardware, into the enable pin of the column driver, but as
these are not wired on the HAD badge, we enable using the row drivers and time using a software loop for short pulses 
and T1 interrupt for longer ones.
At least we have the SPI port to get the data out quicky...

*/

 void interrupt intcode(void)
{
static unsigned int pulsetime=1;
static unsigned char spiptr=0,rowmask=0x7F,bitnum=0; // -> SPI data, ~current row, current BCM bit
unsigned char i;
LATA=0xff; // end previous row enable pulse. Do not add code above here ( or adjust t1offset if you do), as it will mess up timing 

if(++bitnum==nbits) { // all bits done on last row, start new row

  rowmask<<=1;rowmask|=1; // low to enable next row

  if(rowmask==0xff) { // last row done - start of new frame 
    if(newpage) { disppage^=1; newpage=0;} // page swap requested by FG shuffle task
    spiptr=0; // pointer to SPI data
    rowmask=0xfe; // enable first row 
    if(frametimer) frametimer--; // decrement frame-rate timer and hold at zero (for use by FG tasks that want update rate slower than refresh rate)
   
    }  // new frame

// do the first few shortest pulses completely inside int routine due to entry/exit overhead

// macro to send 2 bytes to SPI, for speed. Only need to check status on second due to 1 byte buffer

#define dospi {SSP1BUF=spibuf[disppage][spiptr++];SSP1BUF=spibuf[disppage][spiptr++];while(SSPIF==0);}

  dospi; LATA=rowmask;_delay(bcmtime-1);LATA=0xff; // data for first bit, do first pulse
  dospi; LATA=rowmask;_delay(bcmtime*2-1-2); // -2 to correct time for PORTB read below 
  i=PORTB;   // sample buttons pin during pulse time
  LATA=0xff; // end of pulse
  if(i&0x40) buttons&=rowmask; else buttons|=~rowmask; // set or clear button bit for this row

#if bitsinint>2
  dospi; LATA=rowmask;_delay(bcmtime*4-1);LATA=0xff; // third pulse
#endif
#if bitsinint>3
  dospi; LATA=rowmask;_delay(bcmtime*8-1);LATA=0xff;
#endif
#if bitsinint>4
  dospi; LATA=rowmask;_delay(bcmtime*16-1);LATA=0xff;
#endif

  bitnum=bitsinint;
  pulsetime=(1<<bitsinint)*bcmtime; // calculate time for first pulse to be done using T1 interrupt
  } // new row

else pulsetime*=2; // double pulse time on each subsequent bit

dospi;// send data for next bit

TMR1IF=0;
TMR1=(0xffff+t1offset)-pulsetime; // set timer for pulse length. 
                                  // t1offset compensates for int exit/entry time 
                                  // don't add code after here (or adjust t1offset and make sure code has constant execution time) 
LATA=rowmask; 					  // start pulse
}

//____________________________________________________________________



void shuffle(void)

{ // juggle 8x16 greyscale buffer into SPI BCM data for each pulse period

unsigned char row,spage,spibyte,sptr,dptr,mask;
unsigned int d;


#if presquare==1
// precalculate buffer of gamma-corrected data for speed
for(sptr=0;sptr!=imgbufsize;sptr++) 
{ row=imgbuf[sptr];sqbuf[sptr]=(unsigned int) row*row>>8;}
#endif

spage=disppage?0:1;
sptr=0;
#if bversion==1
for(row=0;row!=8;row+=1)
#else
    for(row=0;row!=64;row+=8)
#endif    
  for(mask=1;mask!=0;mask<<=1) // do each of 8 bits of BCM data

   {
    // macro for speed to generate 1 byte of SPI data for half a column of BCM data
#if bversion==1
    #if presquare==1
       #define tbit(b) if(sqbuf[dptr]&mask) spibyte|=(1<<b); dptr+=8;
    #else
      #define tbit(b) d=imgbuf[dptr];d=(unsigned int) d*d>>8; if(d&mask) spibyte|=(1<<b); dptr+=8;
    #endif
#else
#if presquare==1
       #define tbit(b) if(sqbuf[dptr++]&mask) spibyte|=(1<<b);
    #else
      #define tbit(b) d=imgbuf[dptr++];d=(unsigned int) d*d>>8; if(d&mask) spibyte|=(1<<b);
    #endif
#endif


    spibyte=0;dptr=row+64; tbit(0); tbit(1); tbit(2); tbit(3); tbit(4); tbit(5); tbit(6);  tbit(7); 
    spibuf[spage][sptr++]=spibyte; // second half-column
    
    spibyte=0;dptr=row; tbit(0); tbit(1); tbit(2); tbit(3); tbit(4); tbit(5); tbit(6);  tbit(7); 
    spibuf[spage][sptr++]=spibyte;  // first half-column


    }
newpage=1; // tell int code to page swap at start of next frame

}



//___________________________________________________________________

void fillbuf(char d) 
{char i; for(i=0;i!=imgbufsize;imgbuf[i++]=d);}

void txbyte(char d)
{while(TRMT==0); TXREG=d;}



//=================================================================== effects

#include "sintable.inc"// sine lookup table 8 bit->8 bit

unsigned char sine(unsigned int a)
{ // unsigned char sine of 8:8 fixed point value
return(sintable[a>>8]);
}

signed int ssine(unsigned int a)
{ // signed int sine
return((signed int) sintable[a>>8]-128);
}

//___________________________________________________________________ wave

void wave(void)
{ // wavy 2D sinewave thing
static unsigned int d,xpos,w;
static  int xphase,yphase,yfreq,xfreq,xspeed,yspeed;
static int xp,yp,xf,yf,xs,ys;

unsigned char x,i,a,y;
unsigned int p;


if(start) 
{

xfreq=rand();
yfreq=rand();
xspeed=0x200;
yspeed=0x200;
xphase=rand();
yphase=rand();

}
  
a=0;

for(y=0;y!=16;y++) {

for(x=0;x!=8;x++) {

p=sine(x*xfreq+xphase);
p+=sine(y*yfreq+yphase);
p*=2;
if(p>255) p=255;
imgbuf[a++]=(p^255);
}//x
}//y

xphase+=xspeed;
yphase+=yspeed;

xspeed=ssine(xs)*0x08;
yspeed=ssine(ys)*0x08;
xs+=0x21;
ys+=0x1e;

xfreq=ssine(xf)*0x20+0x1000;
yfreq=ssine(yf)*0x20+0x1000;

xf+=0x2f;
yf+=0x13;
}

//______________________________________________ drops

void drops(void)
#define nd 3 // how many
{
static int x[nd],y[nd],xs[nd],ys[nd]; // 8:8 fixed point for smooth velocity
static char dir;
unsigned char i,c,s,t,f;

speed=1;
if(start) {

for(i=0;i!=nd;i++) {
x[i]=0x800;
y[i]=0x1000;
xs[i]=0; //(rand()&0x7f)+0x10;
ys[i]=(rand()&0x3f)+0x10;

}

} 


for(i=0;i!=nd;i++) {
imgbuf[((x[i]>>8)&7)+((y[i]>>8)&15)*8]=0xff;;

x[i]+=xs[i];
y[i]+=ys[i];

if (x[i]>0x7FF || x[i]<0 || y[i]>0xfff || y[i]<0) {
s=(rand() & 0x3f)+0x10;
/*
switch(++dir&3) { // code for all directions - looks better as simple rain though

case 0 : x[i]=0;xs[i]=s/2;ys[i]=0;y[i]=rand()&0xf00;break;
case 1 : y[i]=0;ys[i]=s;xs[i]=0;x[i]=rand()&0x700;break;
case 2 : x[i]=0x700;xs[i]=-s/2;ys[i]=0;y[i]=rand()&0xf00;break;
case 3 : y[i]=0xf00;ys[i]=-s;xs[i]=0;x[i]=rand()&0x700;break;
}//switch
*/

y[i]=0;ys[i]=(rand()&0x3f)+12;
xs[i]=0;
x[i]=rand()&0x700;

}//edge

}//for
for(i=0;i!=imgbufsize;i++)  { c=imgbuf[i]; if(c) imgbuf[i]-=(c>3)?4:c;}


}

//_________________________________ bouncy antialiased 3d-ish ball

 long dist(int x0,  int x1)
{ // distance squared  between  points 
 int d;
d=((x1>x0)?(x1-x0):(x0-x1));
return((long) d*d);
}

#define maxx 0x400 // edges
#define maxy 0x800

#define sx 100 //  nominal speed
#define sy 180 
#define sz 200

#define maxsize 8000
#define minsize 1500
#define zm 255 // randomness on z bounce
#define rm 15 // randomness on xy bounce

void ball(void)
{
unsigned char a;
 int x,y;
 static int xs,ys,dx,dy,size,ds;
long l;

if(start)  {
dx=(rand()&0xff)-0x80;
dy=(rand()&0xff)-0x80;
xs=0;
ys=0;
size=(rand()&0xfff)+minsize;
ds=sz;
}
for(a=0,y=-0x800;y!=0x800;y+=0x100)   for(x=-0x400;x!=0x400;x+=0x100) 
{

// calc distance squared from each pixel to ball centre. 8:8 bit fixed point for sub-pixel resolution
l=(dist(x,xs)+dist(y,ys))/size;

if(l>255) l=255; // clip distance (background)
l=255-l; // low distance = brightest
imgbuf[a++]=l;

}

xs+=dx; ys+=dy; // update position
if(dx<0) if (xs<-maxx)  dx=sx+(rand()&rm); // bouncy bouncy, with a little randomness
if(dx>0) if (xs>maxx) dx=-sx+(rand()&rm);
if(dy<0) if (ys<-maxy) dy=sy+(rand()&rm);
if(dy>0) if (ys>maxy) dy=-sy+(rand()&rm);

size=size+ds;
if(ds>0) if(size>maxsize) ds=-ds+(rand()&zm);
if(ds<0) if(size<minsize) ds=-ds-(rand()&zm);
}

//__________________________________ virus - life-ish thing based on state of surrounding pixels
void virus(void)
{
unsigned char a,x,y,i,d;
unsigned char ib[128];
int l,m;
static unsigned int  cyc;

if(start) cyc=-1;

if((++cyc&1023)==0) for(i=0;i!=30;i++) {// spray some random pixels in occasionally
a=rand()&127;
imgbuf[a]=rand();
if(a) if (rand() &1) imgbuf[a-1]=250;
if(a<127) if (rand() &1) imgbuf[a+1]=250;
if(a>8) if (rand() &1) imgbuf[a-8]=250;
if(a<119) if (rand() &1) imgbuf[a+8]=250;
}


for(a=0,y=0;y!=16;y++) for(x=0;x!=8;x++) {
l=0;
if(x<15) l+=imgbuf[a+1];
if(x) l+=imgbuf[a-1];
if(y<8) l+=imgbuf[a+8];
if(y) l+=imgbuf[a-8];
m=imgbuf[a];
#define  tl 128
#define th 128
#define die 8
#define grow 64

if(l<tl) if(m) m-=((tl-l)/die);
if(l>th) m+=((l-th)/grow);
 if(m<0) m=0;
 if(m>255) m=0;

ib[a++]=m;

}
for(i=0;i!=128;i++) imgbuf[i]=ib[i];

}

void switchoff(void)
{
unsigned char i,j;
 do {
asm("clrwdt");
 // fadeout, just to be flash!
frametimer=1;while(frametimer);
for(i=0,j=0;j!=imgbufsize;j++) if(imgbuf[j]>2) {imgbuf[j]-=2;i=1;} else imgbuf[j]=0;
shuffle();// blank 
} while(i);


i=0;
do {
asm("clrwdt");
_delay(10000);//debounce delay
if(topbut) i++; else i=0;
} while(i<100); // wait button really up, debounced
INTCON=0;
LATA=0xff;// rows off
SSPCON=0; // make sure SPI pins low
LATB5=1; // powerdown
LATB3=0;LATB1=0;
TXSTA=0; 
RCSTA=0; // UART off 


do
{
INT0IF=0;
INT0IE=1;
asm("sleep");
} while(INT0IF==0);

CCPR1L=effect;
asm("reset");


}


//=================================================================================

void main(void)

{
unsigned char i,x,y,offtimer;

INTCON=0; // disable any ints set up by kernel
OSCCON=0b01110011;

OSCCON2=0b00010000;

TRISA=0;
TRISB=0b01000001;
TRISC=0b00000100;// RC2 connected to column
LATA=0xff;
LATB=0;
LATC=0;

ANSELA=ANSELB=ANSELC=0;
RBPU=0;

TXSTA=0b00100100; // UART for debug
RCSTA=0b10000000;
SPBRG1=0;
SPBRGH1=0;

//SPI, Timer 1 & interrupt setup for greyscale display

disppage=0;

SSP1STAT=0b00000000;
SSP1CON1=0b00110000;

T1CON=0b00000001;
T1GCON=0;
TMR1IE=1;  // enable timer 1 int
PEIE=1;

for(i=0;i!=imgbufsize;i++)imgbuf[i]=i*2; // initial greyscale
shuffle();

frametimer=RI?200:0;// greyscale on reset, not wake
effect=RI?2:CCPR1L; // recover previous effect no. (CCPR1L not cleared by reset)
RI=1;
GIE=1;




//_________________ end of greyscale setup

speed=1; // update every refresh frame

start=1;

do {

do asm("clrwdt");while(frametimer);
frametimer=speed;

butpress= (lastbut ^ buttons) & buttons; // buttons chnaged and now down

if(topbut==1) offtimer=0; else if(++offtimer==offtime) switchoff(); // switchoff validation time

lastbut=buttons; // buttons changed and now down, once per render frame for debounce



if(left) effect=0;
if(right) effect=1; 
if(down) effect=2;
if(up) effect=3;
if(butpress)  {
srand(TMR1) ;// shake up randomizer
start=1; // signal effect to initialise when first entered
}


speed=1;
switch(effect)  { 

case 0 :speed=2; wave(); break;// needs speed=2 to avoid tearing issues 
case 1: drops(); break;
case 2: ball(); break;
case 3 : virus();break;
default : effect=0;
}//switch
LATC0=0;
start=0;



shuffle();


} while(1);

 
vecs(); // force inclusion  of dummy vectors, never executed

	}
	